-- string degDecToDegMinSec(number degDec, string latOrLong = "")
-- Takes LatLong coordinates in decimal degrees like -42.002145 and converts
-- them to degrees, minutes, seconds like -42°00'07"
-- With the optional parameter latOrLong set to either "lat" or "long"
-- (case does not matter) an potential algebraic prefix will be omitted
-- and the string will be postfixed with N, S, E or W like 42°00'07"S
function degDecToDegMinSec(degDec, latOrLong)

  -- Remember algebraic sign in case we need it later
  local algSign = ""
  if (degDec < 0) then
    algSign = "-"
  end

  -- Calculate full degrees, minutes and seconds as positive numbers
  local fullDeg = math.floor(math.abs(degDec)) % 360
  local deg     = math.abs(degDec) % 360
  local min     = (deg - fullDeg) * 60
  local fullMin = math.floor(min)
  local sec     = (min - fullMin) * 60
  local fullSec = math.floor(sec)

  -- Format as string with leading zero
  fullDeg = string.format("%02d", fullDeg)
  fullMin = string.format("%02d", fullMin)
  fullSec = string.format("%02d", fullSec)

  -- Concetenate, including degrees ° minutes ' and seconds " characters
  local str = fullDeg .. "°" .. fullMin .. "'" .. fullSec .. '"'

  -- Without "Lat" or "Long", return the string *including* algebraic prefix
  if latOrLong == nil then
    return algSign .. str
  end

  -- With "Lat" or "Long", figure out postfix
  -- Note, if latOrLong matches neither "lat" nor "long", both prefix and
  -- postfix will be empty, so the result could be technically wrong.
  local postfix = ""
  if string.find(string.lower(latOrLong), "lat") then
    postfix = (degDec > 0) and "N" or "S"
  elseif string.find(string.lower(latOrLong), "long") then
    postfix = (degDec > 0) and "E" or "W"
  end

  -- Done, return the result
  return str .. postfix
end

-- number unitGetAlt(mixed unit)
-- Returns unit's altitude in feet.
-- If unit is of type string, the corresponding unit is searched. Otherwise, it
-- should be an object as returned by Unit.getByName()
function unitGetAlt(unit)

  if type(unit) == "string" then
    unit = Unit.getByName(unit)
  end
  
  if type(unit) ~= "table" then
    return ""
  end

  local point = Object.getPoint(unit)
  if point == nil then
    return ""
  end

  return math.ceil(point.y / 0.3048)
end

-- string unitGetLatLong(mixed unit, string seperator = " ")
-- Returns the unit's LatLong coordinates in degrees, minutes, seconds.
-- If unit is of type string, the corresponding unit is searched. Otherwise, it
-- should be an object as returned by Unit.getByName()
function unitGetLatLong(unit, seperator)
  if type(unit) == "string" then
    unit = Unit.getByName(unit)
  end
  
  if type(unit) ~= "table" then
    env.info("unitGetLatLong() unit is of type " .. type(unit))
    return ""
  end

  local point = Object.getPoint(unit)
  if point == nil then
    return ""
  end

  if type(seperator) ~= "string" then
    seperator = " "
  end
    
  n, e = coord.LOtoLL(point)

  return degDecToDegMinSec(n, "lat") ..  seperator .. degDecToDegMinSec(e, "long")
end

-- number bearing, number clockOffset, number distance unitRelationToUnit(mixed unitFrom, mixed unitTo)
-- First return value is compass bearing from unitFrom to unitTo.
--   Range: >= 0, < 360
--   Example: 350
-- Second return value is clock-method bearing from unitFrom to unitTo.
--   Range: >= 1, <= 12
--   Example: 5
-- Third return value is distance between the two units in meters.
--   Example: 2594
--   Note: Divide by 1852 to get distance in nautical miles
--
-- Note that the clock-method bearing depends on unitFrom's current heading.
--
-- Both unitFrom and unitTo can be given as string or as unit.
function unitRelationToUnit (unitFrom, unitTo)

  -- get unitFrom
  if type(unitFrom) == "string" then
    unitFrom = Unit.getByName(unitFrom)
  end

  -- get unitTo
  if type(unitTo) == "string" then
    unitTo = Unit.getByName(unitTo)
  end
  
  -- Check both units
  if type(unitFrom) ~= "table" then
    env.info("unitRelationToUnit() unitFrom is of type " .. type(unitFrom))
    return ""
  end

  if type(unitTo) ~= "table" then
    env.info("unitRelationToUnit() unitTo is of type " .. type(unitTo))
    return ""
  end

  -- Retrieve unitFrom's position as Pos3 so that it includes the unit's world
  -- coordinates *and* the unit's orientation.
  local posFrom = Object.getPosition(unitFrom)
  if posFrom == nil then
    env.info("unitRelationToUnit() problem with posFrom")
    return ""
  end

  -- For unitTo, we only need its position in the world, so a simpler Vec3
  -- coordinate vector is sufficient
  local pointTo = Object.getPoint(unitTo)
  if pointTo == nil then
    env.info("unitRelationToUnit() problem with pointTo")
    return ""
  end

  -- That's really all the heavy trigonometry lifting:
  -- Get the X and Y offset between the two units
  -- (aka legs/sides/catheteses (?) of the triangle)
  -- Translate coordinate systems to standard 2D X/Y while we're at it
  local deltaX = (posFrom.p.z - pointTo.z)
  local deltaY = (posFrom.p.x - pointTo.x)
 
  -- Hail Pythagoras, because as a by-product we can simply get the distance
  -- between the two units with the Pythagorean theorem.
  -- Coordinates are given in meters, so the distance is in meters as well.
  local distance = math.sqrt(deltaX^2 + deltaY^2)
  distance = mathRound(distance)
 
  -- Calculate the compass heading from unitFrom to unitTo (aka bearing):
  local bearing = math.atan2(deltaY, deltaX) -- Arcus tangent gives the angle in radians
  bearing = math.deg(bearing)                -- Convert to degrees (>= 0, < 360)
  bearing = 270 - bearing                    -- Rotate 3 sectors
  bearing = bearing % 360                    -- Limit to >= 0, < 360

  -- Calculate current compass heading for unitFrom.
  -- Note: In the Caucasus map, this heading is off by circa 6 degrees.
  -- This is probably due to magnetic variation and means we have a 1.7% error
  -- rate in our bearing. With Nevada and other terrains, MagVar will most likely
  -- be different, therefore no static correction is applied here.
  local heading = math.atan2(posFrom.x.z, posFrom.x.x) -- Arcus tangent gives the angle in radians
  heading = math.deg(heading)                          -- Convert to degrees, no need to rotate here
  heading = heading % 360                              -- Limit to >= 0, < 360

  -- Calculate offset from current heading to bearing
  local offset = bearing - heading
  offset = offset % 360
  
  -- Convert to clock-method (360 => 12)
  local clockOffset = offset / 30
  clockOffset = mathRound(clockOffset)
  
  if clockOffset == 0 then
    clockOffset = 12
  end

  bearing = mathRound(bearing)
  heading = mathRound(heading)

  -- local str =  "Bearing: " .. bearing .. "\n"
  -- str = str .. "Heading: " .. heading .. "\n"
  -- str = str .. "Offset: "  .. offset  .. "\n"
  -- str = str .. "Clock-Offset: "  .. clockOffset  .. "\n"
  
  return bearing, clockOffset, distance
end

-- Huh, math.round() seems to be missing.
-- mathRound(number x) rounds x to the nearest integer.
-- .5 will be rounded up.
-- Examples:
-- > = mathRound(3)
-- 3
-- > = mathRound(3.1)
-- 3
-- > = mathRound(3.4999)
-- 3
-- > = mathRound(3.5)
-- 4
-- > = mathRound(3.9)
-- 4
function mathRound (number)
  local flt = number - math.floor(number)
  if flt < 0.5 then
    return math.floor(number)
  else
    return math.ceil(number)
  end
end

-- Simple, recursive, rudimentary dump function to view type and contents of
-- arbitrary variables.
function varDump (var, level)

  if level == nil then
    level = ""
  end

  str = level .. type(var) .. ": "

  if type(var) == "string" or type(var) == "number" then
    return str .. var .. "\n"
  end

  str = str .. "\n"
  level = level .. "  "
  if type(var) == "table" then
    for k, v in pairs(var) do
      if type(v) == "string" or type(v) == "number" then
        str = str .. level .. k .. " (" .. type(v) .. ") => " .. v .. "\n"
      elseif type(v) == "table" then
        str = str .. level .. k .. varDump(v, level) .. "\n"
      end     
    end
  end
  
  -- ToDo: Support other types beside string, number and table?
  return str
end